#!/usr/bin/env python2.7
import argparse
import datetime
import gevent
from gevent.monkey import  patch_all; patch_all()
from gevent.pool import Event
import base64
import json
import logging
import os
import sys
import requests
from requests.exceptions import HTTPError, ReadTimeout, ConnectionError
import uuid
from urllib2 import URLError
import asterisk.manager
import odoorpc
from odoorpc.error import RPCError

# Hack for slef-signed certificates.
if os.getenv('VERIFY_CERT') == '0':
    # Monkey patch SSL for self signed certificates
    import ssl
    if hasattr(ssl, '_create_unverified_context'):
        ssl._create_default_https_context = ssl._create_unverified_context
    # Supress InsecureRequestWarning on odoo_bus_poll
    import urllib3
    urllib3.disable_warnings()


logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')
logger = logging.getLogger(__name__)

ORIGINATE_RESPONSE_CODES = {
    '0': 'No such extension / number or bad dial tech ie. name of a SIP trunk'
         'that doesn\'t exist',
    '1': 'No answer',
    '4': 'Answered',
    '5': 'Busy',
    '8': 'Congested or not available (Disconnected Number)'
}


def ami_connection(func):
    def wrapper(agent, *args, **kwargs):
        try:
            res = func(agent, *args, **kwargs)
            return res
        except asterisk.manager.ManagerException as e:
            logger.exception(e)
            if e.message == 'Not connected':
                agent.ami_disconnected.set()
    return wrapper


def get_ami_value(response, option):
    # Used by ami functions to get a value from AMI response.
    value = None
    for line in response:
        o, v = line.strip().split(':')
        if o == option:
            value = v
            logger.debug('Option: {}, value: {}'.format(o, v))
            break
    else:
        logger.debug('Option {} not found in response: {}'.format(
                                                            option, response))
    return value


class Agent(object):
    version = '1.0.9'
    asterisk_host = os.environ.get('ASTERISK_HOST', 'asterisk')
    ami_port = os.environ.get('MANAGER_PORT', '5038')
    ami_user = os.environ.get('MANAGER_LOGIN', 'odoo')
    ami_password = os.environ.get('MANAGER_PASSWORD')
    ami_manager = None
    ami_disconnected = Event()
    ami_connected = Event()
    stopped = Event()
    settings = {}
    greenlets = []
    bus_greenlet = None
    odoo = None
    odoo_connected = Event()
    odoo_disconnected = Event()
    odoo_session = requests.Session()
    originate_context = 'test-users'
    originate_timeout = 30
    active_originate_calls = {}
    active_callback_calls = {}
    notify_uid = None
    last_ping = datetime.datetime.now()


    def __init__(self):
        self.odoo_disconnected.set()
        if os.environ.get('DEBUG', '0') == '1':
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    def spawn(self, func, *args, **kwargs):
        try:
            self.greenlets.append(gevent.spawn(func, *args, **kwargs))
        except Exception as e:
            logger.exception(e)


    def start(self):
        logger.info('Agent version {} starting'.format(self.version))
        self.spawn(self.ami_connection_loop)
        gevent.spawn(self.connect_odoo)
        gevent.spawn(self.get_originate_settings)
        self.bus_greenlet = gevent.spawn(self.odoo_bus_poll)
        gevent.spawn(self.notify_on_start)
        self.spawn(self.bus_watchdog)
        self.spawn(self.ami_watchdog)
        gevent.joinall(self.greenlets)

    def bus_watchdog(self):
        ping_expect_timeout_minutes = float(os.environ.get('PING_EXPECT_TIMEOUT_MINUTES', '1'))
        logger.debug(u'Ping timeout minutes: {}'.format(ping_expect_timeout_minutes))
        gevent.sleep(60 * ping_expect_timeout_minutes)
        while True:
            logger.debug(u'Bus watchdog check')
            ping_diff = datetime.datetime.now() - self.last_ping
            if ping_diff.total_seconds() > 60 * ping_expect_timeout_minutes:
                logger.warning('Watchdog timeout. Reconnecting to Odoo')
                self.odoo_disconnected.set()
                if self.bus_greenlet:
                    self.bus_greenlet.kill()
                    self.bus_greenlet = gevent.spawn(self.odoo_bus_poll)
            gevent.sleep(60 * ping_expect_timeout_minutes)


    def notify_on_start(self):
        if self.notify_uid:
            logger.debug(u'Sending on start notify to UID {}'.format(self.notify_uid))
            self.odoo_connected.wait()
            self.notify_user(self.notify_uid,
                {'Response': 'Agent', 'Message': 'Agent has been started!'})
        else:
            logger.debug(u'Notify on start not set')


    def stop(self):
        self.stopped.set()


    def connect_odoo(self):
        while not self.stopped.is_set():
            try:
                host = os.environ.get('ODOO_HOST', 'odoo')
                port = os.environ.get('ODOO_PORT', '8069')
                protocol = 'jsonrpc+ssl' if os.environ.get(
                    'ODOO_SCHEME') == 'https' else 'jsonrpc'
                logger.info(
                    u'Connecting to Odoo at {}://{}:{}'.format(
                                                        protocol, host, port))
                odoo = odoorpc.ODOO(host, port=port, protocol=protocol)
                odoo.login(
                    os.environ.get('ODOO_DB', 'astcalls'),
                    os.environ.get('ODOO_LOGIN', 'asterisk_agent'),
                    os.environ.get('ODOO_PASSWORD', 'service'),
                )
                logger.info(u'Connected to Odoo as {}'.format(
                                                os.environ.get('ODOO_LOGIN')))
                self.odoo = odoo
                self.odoo_connected.set()
                self.odoo_disconnected.clear()
                self.last_ping = datetime.datetime.now()
                # Generate a channel token and send it to Odoo
                self.channel_token = uuid.uuid4().hex
                self.odoo.env['asterisk_calls.agent'].set_channel_token(
                                                            self.channel_token)
                # Send 1-st message that will be omitted
                self.odoo.env['asterisk_calls.util'].asterisk_send_bus(
                            'asterisk_agent/{}'.format(
                                                self.channel_token),
                            json.dumps({'name': 'ping'}))
                self.odoo_disconnected.wait()

            except RPCError as e:
                if 'res.users()' in str(e):
                    logger.error(
                            'Odoo login {} not found or bad password {}, '
                            'check in Odoo!'.format(
                                os.environ.get('ODOO_LOGIN', 'asterisk_agent'),
                                os.environ.get('ODOO_PASSWORD', 'service')))
                else:
                    logger.exception(e)
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
            except URLError as e:
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
                logger.error(e)
            except Exception as e:
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
                if 'Connection refused' in repr(e):
                    logger.error('Odoo refusing connection.')
                else:
                    logger.exception(e)

            gevent.sleep(float(os.environ.get('ODOO_RECONNECT_TIMEOUT', '1')))


    def get_originate_settings(self):
        self.odoo_connected.wait()
        def _set_param(param):
            value = self.odoo.execute(
                'asterisk_calls.util', 'get_asterisk_calls_param', [param])
            if not value:
                logger.warning(
                    'Param {} not set in Odoo, using default: {}'.format(
                                                param, getattr(self, param)))
            else:
                logger.debug(u'Settings {} to {}'.format(param, value))
                setattr(self, param, value)
        map(_set_param, ['originate_context', 'originate_timeout'])


    def select_db(self):
        if not self.odoo_connected.is_set() or self.odoo_disconnected.is_set():
            logger.debug('Selecting Odoo database (session refresh)')
            auth_url = '{}://{}:{}/web/session/authenticate'.format(
                os.environ.get('ODOO_SCHEME', 'http'),
                os.environ.get('ODOO_HOST', 'odoo'),
                os.environ.get('ODOO_POLLING_PORT', '8072'))
            data = {
                'jsonrpc': '2.0',
                'params': {
                    'context': {},
                    'db': os.environ.get('ODOO_DB', 'astcalls'),
                    'login': os.environ.get('ODOO_LOGIN', 'asterisk_agent'),
                    'password': os.environ.get('ODOO_PASSWORD', 'service'),
                },
            }
            headers = {
                'Content-type': 'application/json'
            }
            #req = Request('POST', url, data=json.dumps(data), headers=headers)
            rep = self.odoo_session.post(auth_url, data=json.dumps(data),
                                         headers=headers)
            result = rep.json()
            if rep.status_code != 200 or result.get('error'):
                logger.error(u'Odoo authenticate error {}: {}'.format(
                                        rep.status_code,
                                        json.dumps(result['error'], indent=2)))
            else:
                logger.info('Odoo authenticated for long polling')

    def odoo_bus_poll(self):
        if os.getenv('DISABLE_ODOO_BUS_POLL') == '1':
            logger.info('Disabling Odoo bus poll')
            return
        self.odoo_connected.wait()
        last = 0
        while not self.stopped.is_set():
            try:
                bus_url = '{}://{}:{}/longpolling/poll'.format(
                                os.environ.get('ODOO_SCHEME', 'http'),
                                os.environ.get('ODOO_HOST', 'odoo'),
                                os.environ.get('ODOO_POLLING_PORT', '8072'))
                logger.debug(u'Starting odoo bus polling at {}'.format(bus_url))
                # Select DB first
                self.select_db()
                # Now let try to poll
                r = self.odoo_session.post(
                            bus_url,
                            verify=bool(int(os.getenv('VERIFY_CERT', '0'))),
                            headers={'Content-Type': 'application/json'},
                            json={
                                'params': {
                                    'last': last,
                                    'channels': ['asterisk_agent/{}'.format(
                                                        self.channel_token)]
                                }})
                result = r.json().get('result')
                if not result:
                    error = r.json().get('error')
                    if error:
                        logger.error(json.dumps(error, indent=2))
                        gevent.sleep(float(os.environ.get(
                                            'ODOO_RECONNECT_TIMEOUT', '1')))
                        continue
                if last == 0:
                    # Ommit queued data
                    for msg in result:
                        message = json.loads(msg['message'])
                        logger.debug(u'Ommit {}'.format(
                            json.dumps(message, indent=2)))
                        last = msg['id']
                    continue

                for msg in result:
                    last = msg['id']
                    message = json.loads(msg['message'])
                    logger.debug(u'Handle {}'.format(json.dumps(message,
                                                                indent=2)))
                    gevent.spawn(self.handle_bus_message,
                                 msg['channel'], msg['message'])
            except ValueError as e:
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
                if 'No JSON object could be decoded' in e.message:
                    logger.error('JSON result error: {}'.format(r.text))
                else:
                    logger.exception(e)
            except ConnectionError as e:
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
                logger.error(u'Odoo connection error: {}'.format(e.message))
            except Exception as e:
                self.odoo_connected.clear()
                self.odoo_disconnected.set()
                logger.exception(e)
            gevent.sleep(float(os.environ.get('ODOO_RECONNECT_TIMEOUT', '1')))



    def handle_bus_message(self, channel, msg):
        self.last_ping = datetime.datetime.now()
        msg = json.loads(msg)
        name = msg.pop('name')
        logger.debug(u'Bus message {}'.format(name))
        if hasattr(self, 'on_bus_{}'.format(name)):
            getattr(self, 'on_bus_{}'.format(name))(msg)
        else:
            logger.error('Bus handler not found for {}'.format(name))


    def notify_user(self, uid, msg, title=None):
        if not uid:
            logger.debug(u'No uid, will not notify')
            return
        if title:
            msg['Response'] = title
        logger.debug(u'Notify user: {}'.format(json.dumps(msg, indent=2)))
        if msg.get('Success'):
            notify_type = 'notify_info_{}'.format(uid)
        else:
            notify_type = 'notify_warning_{}'.format(uid)
        self.odoo.env['asterisk_calls.util'].asterisk_send_bus(
                                     notify_type,
                                     {'message': msg.get('Message', ''),
                                      'title': msg.get('Response', '')})


    def add_cdr_user_field(self, channel, value):
        result = self.ami_manager.send_action({
            'Action': 'Getvar',
            'Channel': channel,
            'Variable': 'CDR(userfield)',
        })
        # Wierd, I get here results from previous actions
        if 'Variable' in result.headers and 'Value' in result.headers:
            if result.headers.get('Response') == 'Success':
                vals = []
                if result.headers['Value']:
                    vals.append(result.headers['Value'])
                vals.append(value)
                self.ami_manager.send_action({
                    'Action': 'Setvar',
                    'Channel': channel,
                    'Variable': 'CDR(userfield)',
                    'Value': ';'.join(vals)
                })
            else:
                logger.warning(u'add_cdr_user_field error: {}'.format(json.dumps(
                                                                result.headers)))

    def on_bus_restart(self, msg):
        logger.info(u'Restarting')
        args = sys.argv[:]
        args.append('--notify_uid={}'.format(msg['uid']))
        os.execv(sys.executable, ['python2.7'] + args)


    def on_bus_ami_send_action(self, msg):
        result = self.ami_manager.send_action(msg)
        self.notify_user(msg['uid'], result.headers)


    def on_bus_setvar(self, msg):
        msg['Action'] = 'Setvar'
        logger.debug(u'Bus event: {}'.format(json.dumps(msg, indent=2)))
        res = self.ami_manager.send_action(msg)


    @ami_connection
    def on_bus_originate(self, msg):
        uid = msg.pop('uid')
        result = {'Response': 'Started', 'Message': 'Call started'}
        # Check if AMI is alive
        if not self.ami_manager:
            logger.error('AMI not connected')
            self.notify_user(uid, {
                             'Response': 'Error',
                             'Message': 'Not connected to Asterisk!'})
            return
        # Check if user's peer is reachable
        if not msg['Channel']:
            # Channel for user is not defined
            logger.error('Channel for UID {} not defined!'.format(uid))
            self.notify_user(uid, {
                             'Response': 'Error',
                             'Message': 'Origination channel is not defined!'})
            return
        if msg['Channel'].upper()[:4] == 'SIP/':
            status = self.ami_manager.send_action({'Action': 'SIPshowpeer',
                                        'Peer': msg['Channel'].split('/')[1]})
            if status.headers['Response'] == 'Error':
                self.notify_user(uid, status.headers)
                logger.error(u'Failed originate for {}: {}'.format(
                                                    msg['Channel'],
                                                    status.headers['Message']))
                return False
            elif not ('ok' in status.headers['Status'].lower() or
                      'unmonitored' in status.headers['Status'].lower()):
                logger.info(
                    'Failed originate for {}, peer status is {}'.format(
                                                    msg['Channel'],
                                                    status.headers['Status']))
                self.notify_user(uid, {
                        'Response': 'Error',
                        'Message': '{} status is {}!'.format(
                                msg['Channel'],
                                status.headers['Status'].lower())})
                return False
            else:
                logger.debug(u'Peer {} status is {}, originating'.format(
                            msg['Channel'], status.headers['Status']))
        elif msg['Channel'].upper()[:6] == 'PJSIP/':
            # For PJSIP we have different status check actions.
            status = self.ami_manager.send_action({
                                    'Action': 'PJSIPShowEndpoint',
                                    'Endpoint': msg['Channel'].split('/')[1]})
            if status.headers['Response'] == 'Error':
                err_msg = dict(
                    [k.strip().split(':') for k in status.response])['Message']
                self.notify_user(uid, {'Response': 'Error',
                                       'Message': err_msg})
                logger.error(u'Failed originate for {}: {}'.format(
                                                    msg['Channel'], err_msg))
                return False
            # We do not check endpoint status as panoramisk is not ready for pjsip yet.
        # Go-go-go kukuruza!
        channel_id = uuid.uuid4().hex
        other_channel_id = uuid.uuid4().hex
        # Check that there is no active calls for user
        logger.debug(u'Active calls: {}'.format(
                            json.dumps(self.active_originate_calls, indent=2)))
        # Find if user alreade has active call
        for _, v in self.active_originate_calls.items():
            if v['exten'] == msg['Exten'] and uid == v['uid']:
                logger.debug(u'Call is in progress')
                result = {'Response': 'Call failed', 'Message': 'Call already in progress'}
                return self.notify_user(uid, result)
        action = {
            'Action': 'Originate',
            'EarlyMedia': 'true',
            'Context': self.originate_context,
            'Priority': '1',
            'Timeout': float(self.originate_timeout) * 1000,
            'OtherChannelId': other_channel_id,
            'ChannelId': channel_id,
            'Async': 'true',
        }
        # Check Variable to be a list of vars
        if msg.get('Variable'):
            if type(msg['Variable']) != list:
                msg['Variable'] = [msg['Variable']]
        else:
            # Initialize with empty list
            msg['Variable'] = []
        # Check for # in number for DTMF post-dial.
        if '#' in msg['Exten']:
            dtmf_digits = msg['Exten'].split('#')[-1]
            # Fun trick - pause before sending DTMF is number of #.
            hash_count = len(msg['Exten'].split('#')) - 1
            msg['Variable'] += ['__dtmf_digits={}'.format(dtmf_digits),
                               '__dtmf_delay={}'.format(hash_count)]
            msg['Exten'] = msg['Exten'].split('#')[0]
        action.update(msg)
        logger.debug(u'Adding call to active calls')
        self.active_originate_calls[channel_id] = {
            'exten': msg['Exten'],
            'uid': uid
        }
        logger.debug('Sending action: {}'.format(json.dumps(msg, indent=2)))
        res = self.ami_manager.send_action(action)
        if str(res) != 'Success':
            self.active_originate_calls.pop(channel_id)
            self.notify_user(uid, res.headers)
        logger.info(u'Originate call for {} to {} status: {}'.format(
            msg['Channel'], msg['Exten'], res))


    def on_bus_originate_callback(self, msg):
        callback_id = msg.get('callback_id')
        cb = self.odoo.env['asterisk_calls.callback'].browse(callback_id)
        queue = msg.get('queue')
        if queue:
            logger.debug('Queue {} is set, check it'.format(queue))
            # Check if queue exists
            res = self.ami_manager.send_action({
               'Action': 'GetVar',
               'Variable': 'QUEUE_EXISTS({})'.format(queue)
            })
            queue_exists = int(get_ami_value(res.response, 'Value'))
            if queue_exists:
                logger.debug('Queue {} exists'.format(queue))
                # Check if Queue is empty
                res = self.ami_manager.send_action({
                   'Action': 'GetVar',
                   'Variable': 'QUEUE_MEMBER({},ready)'.format(queue)
                })
                ready_count = int(get_ami_value(res.response, 'Value'))
                if ready_count > 0:
                    logger.info(
                        'Queue {} is ready for new callback'.format(queue))
                else:
                    logger.info('Queue {} is not ready'.format(queue))
                    # Update callback status and exit.
                    cb.write({'status_description':
                              'Queue is not ready for new callback'})
                    return
            else:
                # Queue is set but does not exist
                logger.error('Queue {} is set but does not '
                             'exist in Asterisk'.format(queue))
                cb.write({'status': 'failed', 'status_description':
                          'Queue {} does not exists in Asterisk'.format(queue)})
                return
        # Originate callback
        channel_id = uuid.uuid4().hex
        other_channel_id = uuid.uuid4().hex
        action = {
            'Action': 'Originate',
            'Channel': msg['channel'],
            'Exten': msg['exten'],
            'CallerID': msg.get('callerid', ''),
            'Context': msg['context'],
            'Priority': '1',
            'Variable': '__CALLBACK_ID={}'.format(callback_id),
            'EarlyMedia': 'true',
            'Async': 'true',
            'Timeout': float(self.originate_timeout) * 1000,
            'OtherChannelId': other_channel_id,
            'ChannelId': channel_id,
        }
        logger.debug(u'Originate callback: {}'.format(
                                            json.dumps(action, indent=2)))
        # Add callback call to the internal dict
        self.active_callback_calls[channel_id] = callback_id
        # Send AMI action
        res = self.ami_manager.send_action(action)


    def on_bus_hangup_request(self, msg):
        self.add_cdr_user_field(msg['channel'],
                                'Hangup by user id {}'.format(msg['uid']))
        result = self.ami_manager.send_action({
            'Action': 'Hangup',
            'Channel': msg['channel'],
            'Cause': msg.get('cause', '16'),
        })
        self.notify_user(msg['uid'], result.headers)


    def on_bus_redirect(self, msg):
        self.add_cdr_user_field(msg['channel'],
                                'Redirect by user id {}'.format(msg['uid']))
        result = self.ami_manager.send_action({
            'Action': 'Redirect',
            'Channel': msg['channel'],
            'Exten': msg['exten'],
            'Context': msg['context'],
            'Priority': '1'
        })
        self.notify_user(msg['uid'], result.headers)


    def on_bus_spy(self, msg):
        self.add_cdr_user_field(msg['spy_channel'],
                                'Spy by user id {}'.format(msg['uid']))
        result = self.ami_manager.send_action({
            'Action': 'Originate',
            'Channel': msg['user_channel'],
            'Application': 'ChanSpy',
            'Data': '{},{}'.format(msg['spy_channel'], msg['spy_options']),
            'Callerid': 'Spy-{} <{}>'.format(msg['spy_exten'], msg['spy_exten']),
            'Variable': '__SIPADDHEADER="{}"'.format(msg['sip_alert_info'] or '')
        })
        self.notify_user(msg['uid'], result.headers)


    def on_bus_mute_audio(self, msg):
        self.add_cdr_user_field(msg['channel'],
                                'Mute {} by user id {}'.format(
                                                               msg['state'],
                                                               msg['uid']))
        result = self.ami_manager.send_action({
            'Action': 'MuteAudio',
            'Channel': msg['channel'],
            'Direction': msg['direction'],
            'State': msg['state']
        })
        if str(result) == 'Success':
            result.headers['Title'] = 'Mute Audio'
            result.headers['Message'] = 'State: {}'.format(msg['state'])
        self.notify_user(msg['uid'], result.headers)


    def on_bus_ping(self, msg):
        logger.debug(u'Ping received')
        self.last_ping = datetime.datetime.now()


    def ami_watchdog(self):
        manager_ping_interval = float(
                                os.environ.get('MANAGER_PING_INTERVAL', '30'))
        logger.debug(u'Manager ping interval: {}'.format(manager_ping_interval))
        while True:
            gevent.sleep(manager_ping_interval)
            if not self.ami_connected.is_set():
                logger.info('Manager not connected')
                continue
            logger.debug(u'Manager ping check')
            try:
                res = self.ami_manager.send_action({'Action': 'Ping'})
                if res.headers.get('Response') != 'Success':
                    raise Exception('Bad response: {}'.format(res.headers))
            except asterisk.manager.ManagerException as e:
                logger.warning('Lost manager connection: {}'.format(e))
                self.ami_connected.clear()
                self.ami_disconnected.set()
            except Exception:
                logger.exception('Manager ping error:')
                self.ami_connected.clear()
                self.ami_disconnected.set()


    def ami_connection_loop(self):
        while True:
            if self.stopped.is_set():
                return
            try:
                # Create PBX connection
                manager = asterisk.manager.Manager()
                logger.debug(u'Connecting to {}:{} with {}:{}'.format(
                    self.asterisk_host, self.ami_port,
                    self.ami_user,
                    self.ami_password[0] + '******' + self.ami_password[-1:]))
                #
                manager.connect(str(self.asterisk_host),
                                port=int(self.ami_port))
                manager.login(self.ami_user, self.ami_password)
                logger.info(u'Connected to AMI')
                # Register for events
                manager.register_event('VarSet', self.on_asterisk_event)
                manager.register_event('Hangup', self.on_asterisk_event)
                manager.register_event('Cdr', self.on_asterisk_event)
                manager.register_event('Newchannel', self.on_asterisk_event)
                manager.register_event('Newstate', self.on_asterisk_event)
                manager.register_event('NewConnectedLine', self.on_asterisk_event)
                manager.register_event('OriginateResponse', self.on_asterisk_event)
                if os.getenv('DISABLE_EVENT_AGENT_CALLED') != '1':
                    manager.register_event('AgentCalled',
                                           self.on_asterisk_event)
                self.ami_manager = manager
                self.ami_disconnected.clear()
                self.ami_connected.set()
                self.ami_disconnected.wait()

            except asterisk.manager.ManagerSocketException as e:
                logger.error("Error connecting to the manager: %s" % e)
            except asterisk.manager.ManagerAuthException as e:
                logger.error("Error logging in to the manager: %s" % e)
            except asterisk.manager.ManagerException as e:
                logger.error("Error: %s" % e)
            except Exception as e:
                logger.exception('AMI error:')

            logger.info(u'Reconnecting AMI.')
            gevent.sleep(float(os.environ.get('AMI_RECONNECT_TIMEOUT', '1')))


    def on_asterisk_event(self, event, manager):
        logger.debug(u'Asterisk event: {}'.format(event.name))
        # We can optionally set exact event handlers
        special_event_handlers = {
            'Newstate': self.on_asterisk_Newstate,
            'NewConnectedLine': self.on_asterisk_Newstate,
        }
        try:
            if event.name in special_event_handlers:
                special_event_handlers[event.name](event, manager)
            elif hasattr(self, 'on_asterisk_{}'.format(event.name)):
                getattr(self, 'on_asterisk_{}'.format(event.name))(event, manager)
            else:
                logger.error(
                    'Asterisk event handler for {} not defined'.format(event.name))
        except Exception as e:
            logger.exception(e)


    def on_asterisk_VarSet(self, event, manager):
        # QoS of CDR
        def update_qos():
            value = event.headers.get('Value')
            pairs = [k for k in value.split(';') if k]
            values = {}
            for pairs in pairs:
                k,v = pairs.split('=')
                values.update({k: v})
            values.update({
                'uniqueid': event.headers.get('Uniqueid'),
                'linkedid': event.headers.get('Linkedid')
            })
            gevent.sleep(float(os.environ.get('QOS_UPDATE_DELAY', '3')))
            logger.debug(u'QoS update: \n{}'.format(json.dumps(
                values, indent=2)))
            self.odoo.execute('asterisk_calls.call', 'update_qos', [values])
        #
        if event.headers.get('Variable') == 'RTPAUDIOQOS':
            logger.debug(u'RTPAUDIOQOS VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
            gevent.spawn(update_qos)
        elif event.headers.get('Variable') == 'RTPAUDIOQOSRTTBRIDGED':
            logger.debug(u'RTPAUDIOQOSRTTBRIDGED VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSLOSSBRIDGED':
            logger.debug(u'RTPAUDIOQOSLOSSBRIDGED VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSJITTERBRIDGED':
            logger.debug(u'RTPAUDIOQOSJITTERBRIDGED VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSBRIDGED':
            logger.debug(u'RTPAUDIOQOSBRIDGED VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSLOSS':
            logger.debug(u'RTPAUDIOQOSLOSS VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSJITTER':
            logger.debug(u'RTPAUDIOQOSJITTER VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        elif event.headers.get('Variable') == 'RTPAUDIOQOSRTT':
            logger.debug(u'RTPAUDIOQOSRTT VarSet: {}'.format(
                                        json.dumps(event.headers, indent=2)))
        else:
            logger.debug(u'VarSet for {} not defined.'.format(
                event.headers.get('Variable')))


    def on_asterisk_Newchannel(self, event, manager):
        if event.headers['Channel'].startswith('Local/'):
            return
        logger.debug(u'Newchannel: {}'.format(json.dumps(
                                            dict(event.headers.items()),
                                            indent=2)))
        get = event.headers.get
        logger.info('New channel from {} to {}'.format(
                                                    get('CallerIDNum'),
                                                    get('Exten')))
        data = {
            'channel': get('Channel'),
            'uniqueid': get('Uniqueid'),
            'linkedid': get('Linkedid'),
            'context': get('Context'),
            'connected_line_num': get('ConnectedLineNum'),
            'connected_line_name': get('ConnectedLineName'),
            'state': get('ChannelState'),
            'state_desc': get('ChannelStateDesc'),
            'exten': get('Exten'),
            'callerid_num': get('CallerIDNum'),
            'callerid_name': get('CallerIDName'),
            'accountcode': get('AccountCode'),
            'priority': get('Priority'),
            'timestamp': get('Timestamp'),
            'system_name': get('SystemName'),
            'language': get('Language'),
            'event': get('Event'),
        }
        self.odoo.env['asterisk_calls.channel'].create(data)


    def on_asterisk_Newstate(self, event, manager):
        if event.headers['Channel'].startswith('Local/'):
            return
        gevent.sleep(0.1)  # Just give small priority for create channel method
        if event.headers['ChannelStateDesc'] in ['Down', 'Busy']:
            # No sense in sending Down / Busy updates as Hangup will follow very soon.
            return
        logger.debug(u'Newstate: {}'.format(json.dumps(
                                            dict(event.headers.items()),
                                            indent=2)))
        # Check if it is a originate state change
        active_call = self.active_originate_calls.get(
                                                    event.headers['Uniqueid'])
        if active_call and event.headers['ChannelStateDesc'] == 'Ringing':
                self.notify_user(active_call['uid'], {
                                                'Response': 'Call Progress',
                                                'Message': 'Ringing...'
                })
        get = event.headers.get
        data = {
            'channel': get('Channel'),
            'uniqueid': get('Uniqueid'),
            'linkedid': get('Linkedid'),
            'context': get('Context'),
            'connected_line_num': get('ConnectedLineNum'),
            'connected_line_name': get('ConnectedLineName'),
            'state': get('ChannelState'),
            'state_desc': get('ChannelStateDesc'),
            'exten': get('Exten'),
            'callerid_num': get('CallerIDNum'),
            'callerid_name': get('CallerIDName'),
            'accountcode': get('AccountCode'),
            'priority': get('Priority'),
            'timestamp': get('Timestamp'),
            'system_name': get('SystemName'),
            'language': get('Language'),
            'event': get('Event'),
        }
        self.odoo.execute('asterisk_calls.channel', 'update_channel_state',
                          [data])


    def on_asterisk_Hangup(self, event, manager):
        if event.headers['Channel'].startswith('Local/'):
            return
        logger.debug(u'Hangup: {}'.format(
            json.dumps(event.headers, indent=2)))
        logger.info('Hangup channel from {} to {}, state {}'.format(
                                            event.headers['CallerIDNum'],
                                            event.headers['Exten'],
                                            event.headers['ChannelStateDesc']))
        # Check if this is user originated call
        channel_id = event.headers['Uniqueid']
        if self.active_originate_calls.get(channel_id):
            chan = self.active_originate_calls.pop(channel_id)
            self.notify_user(chan['uid'], {
                'Response': 'Hangup',
                'Message': event.headers['Cause-txt']
            })
        # Remove channel in Odoo
        self.odoo.execute('asterisk_calls.channel', 'hangup_channel', [
                                                dict(event.headers.items())])


    def on_asterisk_Cdr(self, event, manager):
        def upload_recording():
            call_id = event.headers.get('UniqueID')
            file_path = os.path.join(
                os.environ.get('MONITOR_DIR', '/var/spool/asterisk/monitor'),
                '{}.wav'.format(call_id))
            # Cleanup recording file
            if os.path.exists(file_path) and os.stat(file_path).st_size == 44:
                # Empty file
                logger.debug('Removing stale {}'.format(file_path))
                os.unlink(file_path)
                return
            if not int(os.environ.get('RECORDING_UPLOAD_ENABLED', '0')):
                logger.debug(u'Recordings not enabled, skipping')
                return
            # Prepare to upload call recording to Odoo
            gevent.sleep(float(os.environ.get('RECORDING_UPLOAD_DELAY', '5')))
            logger.debug(u'Sending call recording.')
            if not os.path.exists(file_path):
                logger.debug(u'Recording for callid {} not found.'.format(
                    call_id))
                return
            data = open(file_path, mode='rb').read()
            result = self.odoo.execute('asterisk_calls.call',
                                       'save_call_recording', [{
                                        'uniqueid': call_id,
                                        'data': base64.b64encode(
                                                            data).decode()}])
            if not result:
                logger.warning('Odoo save_call_recording result returned False!')
                if int(os.environ.get('RECORDING_KEEP_FAILED_UPLOAD', '0')):
                    logger.debug(u'Recording {} deleted'.format(file_path))
                    os.unlink(file_path)
                else:
                    logger.debug(u'Recording {} not deleted'.format(file_path))
            else:
                logger.debug(u'Call recording saved.')
                if os.getenv('RECORDING_KEEP_AFTER_UPLOAD') != '1':
                    if os.path.exists(file_path):
                        os.unlink(file_path)
                        logger.debug(u'Recording {} deleted'.format(file_path))
                    else:
                        logger.debug(
                            u'Recording {} was already removed'.format(
                                                                    file_path))
                else:
                    logger.debug(u'Recording {} not deleted'.format(file_path))

        # Start of on_asterisk_Cdr
        logger.debug(u'Asterisk event Cdr: {}'.format(json.dumps(
                                            dict(event.headers.items()),
                                            indent=2)))
        get = event.headers.get
        data = {
            'accountcode': get('AccountCode'),
            'src': get('Source'),
            'dst': get('Destination'),
            'dcontext': get('DestinationContext'),
            'clid': get('CallerID'),
            'channel': get('Channel'),
            'dstchannel': get('DestinationChannel'),
            'lastapp': get('LastApplication'),
            'lastdata': get('LastData'),
            'started': get('StartTime') or False,
            'answered': get('AnswerTime') or False,
            'ended': get('EndTime') or False,
            'duration': get('Duration'),
            'billsec': get('BillableSeconds'),
            'disposition': get('Disposition'),
            'amaflags': get('AMAFlags'),
            'uniqueid': get('UniqueID'),
            'linkedid': get('Linkedid'),
            'userfield': get('UserField'),
        }
        # Check if this is a callback call
        if get('UniqueID') in self.active_callback_calls:
            callback_id = self.active_callback_calls.pop(
                                                    event.headers['UniqueID'])
            logger.debug('Callback CDR')
            data['callback'] = callback_id
            cb = self.odoo.env['asterisk_calls.callback'].browse(callback_id)
            if get('Disposition') == 'ANSWERED':
                cb.write({'status': 'done',
                          'status_description': ''})
        # Create CDR
        self.odoo.env['asterisk_calls.call'].create(data, context={
                                             'mail_create_nolog': True,
                                             'mail_create_nosubscribe': True,
                                             'tracking_disable': True,
                                             'mail_notrack': True})
        gevent.spawn(upload_recording)


    def on_asterisk_OriginateResponse(self, event, manager):
        logger.debug(u'Asterisk event OriginateResponse: {}'.format(json.dumps(
                                        dict(event.headers.items()), indent=2)))
        response = event.headers['Response']
        reason = event.headers['Reason']
        channel_id = event.headers['Uniqueid']
        # Check ordinary calls
        if response == 'Failure' and channel_id in self.active_originate_calls:
            logger.debug('Failed originate')
            chan = self.active_originate_calls.pop(channel_id)
            self.notify_user(chan['uid'], {
                'Response': 'Error',
                'Message': ORIGINATE_RESPONSE_CODES.get(
                                                    reason, 'Unknown error')})
        # Check if this is a callback call
        if response == 'Failure' and event.headers.get(
                'Uniqueid') in self.active_callback_calls and reason == '0':
            callback_id = self.active_callback_calls.pop(
                                                    event.headers['Uniqueid'])
            logger.debug('Callback Response: {}'.format(response))
            cb = self.odoo.env[
                'asterisk_calls.callback'].browse(callback_id)
            cb.write({'status': 'failed',
                      'status_description': ORIGINATE_RESPONSE_CODES.get(
                                reason, 'Unknown error')})


    def on_asterisk_AgentCalled(self, event, manager):
        logger.debug(u'Asterisk event AgentCalled: {}'.format(json.dumps(
                                    dict(event.headers.items()), indent=2)))
        msg = {
            'channel': event.headers['Interface'],
            'member_name': event.headers['MemberName'],
            'queue': event.headers['Queue'],
            'event': 'agent_called',
            'dst': event.headers['Exten'],
        }
        self.odoo.env['asterisk_calls.util'].asterisk_send_bus(
                                    'asterisk_calls_channels', json.dumps(msg))


if __name__ == '__main__':
    agent = Agent()
    # Parse args
    parser = argparse.ArgumentParser(description='Asterisk agent')
    parser.add_argument('--notify_uid', type=int,
                        help='Odoo UID to notify on agent start')
    args = parser.parse_args()
    if args.notify_uid:
        agent.notify_uid = args.notify_uid
    # Start!
    agent.start()
